"""Here is my module docstring
"""

import socket
import struct
import logging
import logging.config

_LIBRARY_VERSION = '1.0.0.0'
COMMAND_PORT = 8080
DEFAULT_TIMEOUT_S = 10
RECV_BUFFER_SIZE = 4096
SCAN_RATES_HZ = {
    0: 1000,
    1: 100,
    2: 10,
    3: 1}

# Define logger for this module
logger = logging.getLogger(__name__)


def configure_logger():

    logging_config = dict(
        version=1,
        formatters={
            'f': {
                'format': '%(asctime)s,%(msecs)d %(levelname)-s [%(filename)s:%(lineno)d - %(funcName)s()] %(message)s'
            },
        },
        handlers={
            'h': {'class': 'logging.StreamHandler',
                  'formatter': 'f',
                  'level': logging.DEBUG}
        },
        loggers={
            'hyperion': {
                'handlers': ['h'],
                'level': logging.DEBUG,
                'propagate': False},
            __name__: {
                'handlers': ['h'],
                'level': logging.DEBUG,
                'propagate': False}}
    )

    logging.config.dictConfig(logging_config)

    logger.info('Logger configured.')


class TunableFilterDevelopmentKit:

    def __init__(self, destination_ip_address, local_ip_address='',
                 port=COMMAND_PORT, timeout=DEFAULT_TIMEOUT_S):
        self._destination_ip_address = destination_ip_address
        self._local_ip_address = local_ip_address
        self._port = port
        self._timeout = timeout
        self._initialized = False
        self.initialize_socket()

    def initialize_socket(self):
        if not self._initialized:
            self._commSocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self._commSocket.settimeout(self._timeout)
            self._commSocket.bind((self._local_ip_address, self._port))
            self._initialized = True

            self.get_version()
            self.get_bias()
            self.get_amplitude()
            self.get_scan_rate()

            logger.info('Initialized to Micron Optics Tunable Filter Development Kit at {0}:{1}.'.format(self._destination_ip_address, self._port))
            logger.info('FPGA Version = v{0}'.format(self.version_fpga))
            logger.info('Firmware Version = v{0}'.format(self.version_fw))
            logger.info('Current bias = {0}'.format(self.bias))
            logger.info('Current amplitude = {0}'.format(self.amplitude))
            logger.info('Current scan rate = {0} Hz'.format(self.scan_rate_hz))

        else:
            print('Connection already initialized.')

    def close(self):
        try:
            self._commSocket.close()
        except:
            logger.error('Error in closing down socket.')
        self._initialized = False

    def _execute_command(self, command, arguments=''):
        logger.info(('Executing: {0} {1}'.format(command, arguments)))
        self._write_command(command, arguments)
        self._read_response()

    def _write_command(self, command, arguments):

        command_length = len(command)
        argument_length = len(arguments)
        # Create packet as big endian (network order)
        # 1 byte - reserved
        # 1 byte - reserved
        # 2 bytes - length of command (n)
        # 4 bytes - length or arguments (m)
        # n bytes - command
        # m bytes - arguments
        self._message_out = struct.pack('!BBHI{0}s{1}s'.format(command_length, argument_length),
                                        0,
                                        0,
                                        command_length,
                                        argument_length,
                                        bytes(command, 'utf-8'),
                                        bytes(arguments, 'utf-8'))
        logger.info('Sending {0} to IP {1} port {2}'.format(':'.join('{:02x}'.format(x) for x in self._message_out),
                                                      self._destination_ip_address,
                                                      self._port))

        self._commSocket.sendto(self._message_out, (self._destination_ip_address, self._port))

    def _read_response(self):
        self._response_data = self._commSocket.recv(RECV_BUFFER_SIZE)

        logger.info('Receiving {0} from IP {1} port {2}'.format(':'.join('{:02x}'.format(x) for x in self._response_data),
                                                          self._destination_ip_address,
                                                          self._port))

        header_info = struct.unpack('!BBHI', self._response_data[0:8])

        self._response_status = header_info[0]
        self._response_request_options_echo = header_info[1]
        self._response_message_length = header_info[2]
        self._response_content_length = header_info[3]

        logger.info('Response length = {0}'.format(len(self._response_data)))
        logger.info('Response status = {0}'.format(self._response_status))
        logger.info('Response options = {0}'.format(self._response_request_options_echo))
        logger.info('Response message length = {0}'.format(self._response_message_length))
        logger.info('Response content length = {0}'.format(self._response_content_length))

        data = struct.unpack('!{0}s{1}s'.format(self._response_message_length, self._response_content_length), self._response_data[8:])
        self._response_message = data[0]
        self._response_content = data[1]

        logger.info('Response message = {0}'.format(self._response_message))
        logger.info('Response content = {0}'.format(self._response_content))

    def get_bias(self):
        self._execute_command('#getBias')
        self.bias = int(self._response_content.decode(), 16)

    def get_amplitude(self):
        self._execute_command('#getAmplitude')
        self.amplitude = int(self._response_content.decode(), 16)

    def get_scan_rate(self):
        self._execute_command('#getScanRate')
        self.scan_rate_hz = SCAN_RATES_HZ[struct.unpack('!B', self._response_content[0:1])[0]]

    def get_version(self):
        self._execute_command('#getVersion')

        version_fpga = '{0:d}.{1:d}.{2:d}.{3:d}'.format(int(self._response_content[0:2].decode(), 16),
                                                               int(self._response_content[2:4].decode(), 16),
                                                               int(self._response_content[4:6].decode(), 16),
                                                               int(self._response_content[6:8].decode(), 16))
        version_fw = '{0:d}.{1:d}'.format(int(self._response_content[8:9].decode()),
                                          int(self._response_content[9:10].decode()))

        self.version_fpga = version_fpga
        self.version_fw = version_fw

        return (version_fpga, version_fw)

    def set_bias(self, bias):
        self._execute_command('#setBias', '{0:08x}'.format(bias))

    def set_amplitude(self, amplitude):
        self._execute_command('#setAmplitude', '{0:08x}'.format(amplitude))

    def set_scan_rate(self, scan_rate):
        if scan_rate not in SCAN_RATES_HZ.values():
            logger.error('Not a valid scan rate.')
            print('Not a valid scan rate.')
        else:
            for key in SCAN_RATES_HZ.keys():
                if SCAN_RATES_HZ[key] == scan_rate:
                    self._execute_command('#setScanRate', '{0}'.format(key))
                    break
